#include <map>
#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <functional>
#include <cassert>
#include "read_word.h"

using namespace std;

typedef map<string,unsigned> word_count_t;

void count_words(word_count_t& word_count)
{
   string word;
   while(read_word(word))
      word_count[word]++;
}

struct count_cmp
{
   const word_count_t& word_count;
   count_cmp(const word_count_t& w) : word_count(w) {}

   // comparision is reversed to make descending list
   bool lt(const string *a,const string *b) const {
      return word_count.at(*a) > word_count.at(*b);
   }
   bool gt(const string *a,const string *b) const {
      return word_count.at(*a) < word_count.at(*b);
   }
};

void select1(vector<const string*>& v,int l,int u,int n,const count_cmp& cmp)
{
   if(u-l<n || n==0)
      return;

   swap(v[l],v[l+1+random()%(u-l)]);

   int i=l;
   int j=u+1;
   while(i<=j) {
      do { i++; } while(i<j && cmp.lt(v[i],v[l]));
      do { j--; } while(cmp.gt(v[j],v[l]));
      if(i<j)
	 swap(v[i],v[j]);
   }
   swap(v[l],v[j]);

   if(j-l>=n)
      select1(v,l,j-1,n,cmp);
   else
      select1(v,j+1,u,n-(j+1-l),cmp);
}

void print_top_words(const word_count_t& word_count,unsigned K)
{
   if(word_count.size()<K)
      K=word_count.size();

   vector<const string*> words;
   word_count_t::const_iterator wc;
   for(wc=word_count.begin(); wc!=word_count.end(); ++wc)
	words.push_back(&wc->first);

   select1(words,0,words.size()-1,K,count_cmp(word_count));

   vector<const string*>::iterator word=words.begin();
   for(int i=0; i<K; ++word, ++i)
      cout << **word << '\t' << word_count.at(**word) << endl;
}

int main()
{
   cin.unsetf(ios_base::skipws);

   word_count_t word_count;
   count_words(word_count);

   print_top_words(word_count,100);

   return 0;
}
